///////////////////////////////////////////////////////////////////////////////
// PureCrypt v1.0a                                                           //
///////////////////////////////////////////////////////////////////////////////
// SHA-1 hash routines.                                                      //
///////////////////////////////////////////////////////////////////////////////

// The usual suspects
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
// Endian specific macros - converting to a big endian machine would need to be
// done here
#include "System.h"
// Bits and pieces relevant for the functions below
#include "SHA-1.h"
// Mostly just needed for the HeapAlloc() function
#include <windows.h>

// The PB memory heap handle
extern int *PB_MemoryBase;

// PureCrypt API //////////////////////////////////////////////////////////////

// Returns a 20 byte buffer with the result
PBCHAR PB_SHA1Hash(uchar8 *buffer, ulong32 length)
{
  SHA1 md;
  char *result = NULL;

  // Have we got input?
  if (buffer && *buffer) {
    if (sha1_init(&md) != TRUE) {
      return result;
    }
    if (sha1_process(&md, buffer, length) != TRUE) {
      return result;
    }
    // Allocate some memory within the PB heap
    result = HeapAlloc(PB_MemoryBase, HEAP_ZERO_MEMORY, 20);
    if (sha1_done(&md, (unsigned char *)result) != TRUE) {
      // Free the allocated memory
      HeapFree(PB_MemoryBase, 0, result);
      result = NULL;
      return result;
    }
  }

  return result;
}

// SHA1 Functions /////////////////////////////////////////////////////////////

int  sha1_compress(SHA1 *md, uchar8 *buf)
{
  ulong32 a,b,c,d,e,W[80],i;

  /* copy the state into 512-bits into W[0..15] */
  for (i = 0; i < 16; i++) {
    LOAD32H(W[i], buf + (4*i));
  }

  /* copy state */
  a = md->state[0];
  b = md->state[1];
  c = md->state[2];
  d = md->state[3];
  e = md->state[4];

  /* expand it */
  for (i = 16; i < 80; i++) {
    W[i] = ROL(W[i-3] ^ W[i-8] ^ W[i-14] ^ W[i-16], 1); 
  }

  /* compress */
#ifdef F0
#undef F0
#endif
#define F0(x,y,z)  (z ^ (x & (y ^ z)))
#ifdef FF0
#undef FF0
#endif
#define FF0(a,b,c,d,e,i) e = (ROLc(a, 5) + F0(b,c,d) + e + W[i] + 0x5a827999UL); b = ROLc(b, 30);

  /* round one */
  for (i = 0; i < 20; ) {
    FF0(a,b,c,d,e,i++);
    FF0(e,a,b,c,d,i++);
    FF0(d,e,a,b,c,i++);
    FF0(c,d,e,a,b,i++);
    FF0(b,c,d,e,a,i++);
  }

#ifdef F1
#undef F1
#endif
#define F1(x,y,z)  (x ^ y ^ z)
#ifdef FF1
#undef FF1
#endif
#define FF1(a,b,c,d,e,i) e = (ROLc(a, 5) + F1(b,c,d) + e + W[i] + 0x6ed9eba1UL); b = ROLc(b, 30);

  /* round two */
  for (; i < 40; )  { 
    FF1(a,b,c,d,e,i++);
    FF1(e,a,b,c,d,i++);
    FF1(d,e,a,b,c,i++);
    FF1(c,d,e,a,b,i++);
    FF1(b,c,d,e,a,i++);
  }

#ifdef F2
#undef F2
#endif
#define F2(x,y,z)  ((x & y) | (z & (x | y)))
#ifdef FF2
#undef FF2
#endif
#define FF2(a,b,c,d,e,i) e = (ROLc(a, 5) + F2(b,c,d) + e + W[i] + 0x8f1bbcdcUL); b = ROLc(b, 30);

  /* round three */
  for (; i < 60; )  { 
    FF2(a,b,c,d,e,i++);
    FF2(e,a,b,c,d,i++);
    FF2(d,e,a,b,c,i++);
    FF2(c,d,e,a,b,i++);
    FF2(b,c,d,e,a,i++);
  }

#ifdef F3
#undef F3
#endif
#define F3(x,y,z)  (x ^ y ^ z)
#ifdef FF3
#undef FF3
#endif
#define FF3(a,b,c,d,e,i) e = (ROLc(a, 5) + F3(b,c,d) + e + W[i] + 0xca62c1d6UL); b = ROLc(b, 30);

  /* round four */
  for (; i < 80; )  { 
    FF3(a,b,c,d,e,i++);
    FF3(e,a,b,c,d,i++);
    FF3(d,e,a,b,c,i++);
    FF3(c,d,e,a,b,i++);
    FF3(b,c,d,e,a,i++);
  }

  /* store */
  md->state[0] = md->state[0] + a;
  md->state[1] = md->state[1] + b;
  md->state[2] = md->state[2] + c;
  md->state[3] = md->state[3] + d;
  md->state[4] = md->state[4] + e;

  return TRUE;
}

int sha1_init(SHA1 *md)
{
  int i;

  if (md == NULL) return FALSE;
  md->state[0] = 0x67452301UL;
  md->state[1] = 0xefcdab89UL;
  md->state[2] = 0x98badcfeUL;
  md->state[3] = 0x10325476UL;
  md->state[4] = 0xc3d2e1f0UL;
  md->curlen = 0;
  md->length = 0;
  for (i = 0; i < 64; i++)
    md->buf[i] = ' ';
  return TRUE;
}

int sha1_process(SHA1 *md, uchar8 *in, ulong32 inlen)
{
  ulong32 n;
  int blocksize = 64;

  // Check to see we have been passed valid arguments
  if (md == NULL || in == NULL) {
    return FALSE;
  }
  if (md->curlen > sizeof(md->buf)) {
    return FALSE;
  }

  // Process block
  while (inlen > 0) {
    if (md->curlen == 0 && inlen >= blocksize) {
      if ((sha1_compress(md, (unsigned char *)in)) != TRUE) {
        return FALSE;
      }
      md->length += blocksize * 8;
      in += blocksize;
      inlen -= blocksize;
    } else {
      n = MIN(inlen, (blocksize - md->curlen));
      memcpy(md->buf + md->curlen, in, (size_t)n);
      md->curlen += n;
      in += n;
      inlen -= n;
      if (md->curlen == blocksize) {
        if ((sha1_compress(md, md->buf)) != TRUE) {
          return FALSE;
        }
        md->length += 8*blocksize;
        md->curlen = 0;
      }
    }
  }
  return TRUE;
}

int sha1_done(SHA1 *md, uchar8 *out)
{
  int i;

  if (md == NULL || out == NULL) {
    return FALSE;
  }

  if (md->curlen >= sizeof(md->buf)) {
    return FALSE;
  }

  /* increase the length of the message */
  md->length += md->curlen * 8;

  /* append the '1' bit */
  md->buf[md->curlen++] = (unsigned char)0x80;

  /* if the length is currently above 56 bytes we append zeros
   * then compress.  Then we can fall back to padding zeros and length
   * encoding like normal.
   */
  if (md->curlen > 56) {
    while (md->curlen < 64) {
      md->buf[md->curlen++] = (unsigned char)0;
    }
    sha1_compress(md, md->buf);
    md->curlen = 0;
  }

  /* pad upto 56 bytes of zeroes */
  while (md->curlen < 56) {
    md->buf[md->curlen++] = (unsigned char)0;
  }

  /* store length */
  STORE64H(md->length, md->buf+56);
  sha1_compress(md, md->buf);

  /* copy output */
  for (i = 0; i < 5; i++) {
    STORE32H(md->state[i], out+(4*i));
  }
  return TRUE;
}
