///////////////////////////////////////////////////////////////////////////////
// PureCrypt v1.0a                                                           //
///////////////////////////////////////////////////////////////////////////////
// System specific types and macros.                                         //
///////////////////////////////////////////////////////////////////////////////

// Types and Defines //////////////////////////////////////////////////////////

typedef unsigned long long ulong64;
typedef signed long long   long64;
typedef unsigned long ulong32;
typedef signed long long32;
typedef unsigned char uchar8;

// Macros /////////////////////////////////////////////////////////////////////

// PureBASIC related defines
#define PBINT extern int _stdcall
#define PBLONG extern long32 _stdcall
#define PBCHAR extern char _stdcall *
#define PBFLOAT extern float _stdcall

// Remember that these are set for "little endian" machines
#define CONST64(n) n ## ULL

#define LOAD32H(x, y)                            \
     { x = ((ulong32)((y)[0] & 255)<<24) | \
           ((ulong32)((y)[1] & 255)<<16) | \
           ((ulong32)((y)[2] & 255)<<8)  | \
           ((ulong32)((y)[3] & 255)); }

#define LOAD32L(x, y)         \
     memcpy(&(x), y, 4);

#define STORE32H(x, y)                                                                     \
     { (y)[0] = (uchar8)(((x)>>24)&255); (y)[1] = (uchar8)(((x)>>16)&255);   \
       (y)[2] = (uchar8)(((x)>>8)&255); (y)[3] = (uchar8)((x)&255); }

#define STORE32L(x, y)        \
     { unsigned long __t = (x); memcpy(y, &__t, 4); }

#define LOAD64H(x, y)                                                      \
   { x = (((ulong64)((y)[0] & 255))<<56)|(((ulong64)((y)[1] & 255))<<48) | \
         (((ulong64)((y)[2] & 255))<<40)|(((ulong64)((y)[3] & 255))<<32) | \
         (((ulong64)((y)[4] & 255))<<24)|(((ulong64)((y)[5] & 255))<<16) | \
         (((ulong64)((y)[6] & 255))<<8)|(((ulong64)((y)[7] & 255))); }

#define STORE64H(x, y)                                                                     \
   { (y)[0] = (uchar8)(((x)>>56)&255); (y)[1] = (uchar8)(((x)>>48)&255);     \
     (y)[2] = (uchar8)(((x)>>40)&255); (y)[3] = (uchar8)(((x)>>32)&255);     \
     (y)[4] = (uchar8)(((x)>>24)&255); (y)[5] = (uchar8)(((x)>>16)&255);     \
     (y)[6] = (uchar8)(((x)>>8)&255); (y)[7] = (uchar8)((x)&255); }

#define ROL(x, y) ( (((unsigned long)(x)<<(unsigned long)((y)&31)) | (((unsigned long)(x)&0xFFFFFFFFUL)>>(unsigned long)(32-((y)&31)))) & 0xFFFFFFFFUL)
#define ROR(x, y) ( ((((unsigned long)(x)&0xFFFFFFFFUL)>>(unsigned long)((y)&31)) | ((unsigned long)(x)<<(unsigned long)(32-((y)&31)))) & 0xFFFFFFFFUL)
#define ROLc(x, y) ( (((unsigned long)(x)<<(unsigned long)((y)&31)) | (((unsigned long)(x)&0xFFFFFFFFUL)>>(unsigned long)(32-((y)&31)))) & 0xFFFFFFFFUL)
#define RORc(x, y) ( ((((unsigned long)(x)&0xFFFFFFFFUL)>>(unsigned long)((y)&31)) | ((unsigned long)(x)<<(unsigned long)(32-((y)&31)))) & 0xFFFFFFFFUL)

#define ROL64(x, y) \
    ( (((x)<<((ulong64)(y)&63)) | \
      (((x)&CONST64(0xFFFFFFFFFFFFFFFF))>>((ulong64)64-((y)&63)))) & CONST64(0xFFFFFFFFFFFFFFFF))

#define ROR64(x, y) \
    ( ((((x)&CONST64(0xFFFFFFFFFFFFFFFF))>>((ulong64)(y)&CONST64(63))) | \
      ((x)<<((ulong64)(64-((y)&CONST64(63)))))) & CONST64(0xFFFFFFFFFFFFFFFF))

#define ROL64c(x, y) \
    ( (((x)<<((ulong64)(y)&63)) | \
      (((x)&CONST64(0xFFFFFFFFFFFFFFFF))>>((ulong64)64-((y)&63)))) & CONST64(0xFFFFFFFFFFFFFFFF))

#define ROR64c(x, y) \
    ( ((((x)&CONST64(0xFFFFFFFFFFFFFFFF))>>((ulong64)(y)&CONST64(63))) | \
      ((x)<<((ulong64)(64-((y)&CONST64(63)))))) & CONST64(0xFFFFFFFFFFFFFFFF))

#define byte(x, n) (((x) >> (8 * (n))) & 255)

#undef MAX
#undef MIN
#define MAX(x, y) ( ((x)>(y))?(x):(y) )
#define MIN(x, y) ( ((x)<(y))?(x):(y) )

#define TRUE 1
#define FALSE 0
